/*
 * Copyright (c) 1999, 2000, 2001, 2002, 2003, Frye Electronics
 *
 * S1D13505 (ex-1355) driver for Frye Fonix 7000
 * S1D13506 (ex-1356)
 *
 * Original by Kaben Nanlohy <kaben@frye.com>
 * hacked on by Aaron Grier <aaron@frye.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h> /* memset and memmove */

#include "scr_fryebox.h"

#include "device.h"
#include "genfont.h"
#include "genmem.h"
#include "fb.h"

/* default screen dimensions for the 7k are defined via modeline.  depth
 * is defined here.  */
#ifndef SCREEN_DEPTH
#define SCREEN_DEPTH 8
#endif

/* and pixel formats for the various bitdepths */
#ifndef MWPIXEL_FORMAT
#if SCREEN_DEPTH <= 8
#define MWPIXEL_FORMAT MWPF_PALETTE
#elif SCREEN_DEPTH == 16
#define MWPIXEL_FORMAT MWPF_TRUECOLOR565
#elif SCREEN_DEPTH == 24
#define MWPIXEL_FORMAT MWPF_TRUECOLOR888
#elif SCREEN_DEPTH == 32
#define MWPIXEL_FORMAT MWPF_TRUECOLOR0888
#else
#error "bad screen depth"
#endif
#endif

/*
 * video mode selection
 * analog-interfaced LCDs are notoriously finicky with timing.  best to
 * stick with VESA standards when possible.
 */

/* VESA 640x480 @ 60Hz with 25.175MHz dot clock */
#define MODELINE25175	"25.175 640 664 760 800  480 491 493 525 -h -v"

/* VESA 640x480 @ 72Hz with 31.500MHz dot clock */
#define MODELINE31500	"31.500 640 664 704 832  480 489 492 520 -h -v"

/* hack-y 640x480 @ 75Hz with 32.000MHz dot clock */
#define MODELINE32000	"32.00  640 664 704 834  480 489 491 511 -h -v"

/* VESA 800x600 @ 60Hz with 40.000MHz dot clock
 * this is as high as the epson controller can go! */
#define MODELINE40000 "40.000 800 840 968 1056  600 601 605 628 +h +v"

#if 0
/* 640x480 @ with 40.000MHz dot clock won't happen.  */
#define DEFAULTMODELINE "40.000 640 704 832 904  480 530 542 615 +h -v"
#endif

/* debugging infos */
#if 1
#define DEBUG_SCR
static char doubledebug = 0;
#define dprintf printf
#define ddprintf if (doubledebug) dprintf
#define ddon doubledebug = 1
#define ddoff doubledebug = 0
#else
#define dprintf(ignored...)
#define ddprintf(ignored...)
#define ddon
#define ddoff
#endif

/* copyright and version information */
/* $Format: "const char _MW_Frye_version[]=\"$ProjectVersion$\";"$ */
const char _MW_Frye_version[]="nanosteve.1";

/* FB7k driver entry points*/
static PSD  FB7k_open(PSD psd);
static void FB7k_close(PSD psd);
static void FB7k_getscreeninfo(PSD psd,PMWSCREENINFO psi);;
static void FB7k505_setpalette(PSD psd,int first,int count,MWPALENTRY *pal);
static void FB7k506_setpalette(PSD psd,int first,int count,MWPALENTRY *pal);
static int
FB7k_init(PSD psd);
static void
FB7k_drawpixel(PSD psd, MWCOORD x, MWCOORD y, MWPIXELVAL c);
static MWPIXELVAL
FB7k_readpixel(PSD psd, MWCOORD x, MWCOORD y);
static void
FB7k_drawhorzline(PSD psd, MWCOORD x1, MWCOORD x2, MWCOORD y, MWPIXELVAL c);
static void
FB7k_drawvertline(PSD psd, MWCOORD x, MWCOORD y1, MWCOORD y2, MWPIXELVAL c);
static void
FB7k_drawfillrect(PSD psd,MWCOORD x1, MWCOORD y1, MWCOORD x2,
    MWCOORD y2, MWPIXELVAL c);
static void
FB7k_blit(PSD dstpsd, MWCOORD dstx, MWCOORD dsty, MWCOORD w, MWCOORD h,
    PSD srcpsd, MWCOORD srcx, MWCOORD srcy, long op);
static void
FB7k_stretchblit(PSD dstpsd, MWCOORD dstx, MWCOORD dsty, MWCOORD dstw,
    MWCOORD dsth, PSD srcpsd, MWCOORD srcx, MWCOORD srcy, MWCOORD srcw,
    MWCOORD srch, long op);
static void     
FB7k_conditioned_line(PSD psd,
    MWCOORD x1, MWCOORD y1, MWCOORD x2, MWCOORD y2,
    MWCOORD dx, MWCOORD dy, MWCOORD xi, MWCOORD yi,
    MWCOORD r, MWPIXELVAL c);
static MWBOOL
FB7k_mapmemgc(PSD mempsd,MWCOORD w,MWCOORD h,int planes,int bpp,int linelen, int size, void *addr);
static void
FB7k_DrawText(PSD psd,MWCOORD x,MWCOORD y, const MWUCHAR *str, int count,
	MWPIXELVAL fg, PMWFONT pfont); 
static void
FB7kBitBLT_DrawText(PSD psd,MWCOORD x,MWCOORD y, const MWUCHAR *str, int count,
	MWPIXELVAL fg, PMWFONT pfont); 

/* variables */
static const int FB7k_pal_max = (1 << SCREEN_DEPTH);
SCREENDEVICE    scrdev = {
    0, 0,               /* xres, yres -- X, Y screen res (real) */
    0, 0,               /* xvirtres, yvirtres -- X, Y drawing res */
    
    0,                  /* planes */
    0,                  /* bpp */
    0,                  /* linelen -- line length
                         * in bytes for bpp 1, 2, 4, 8
                         * in pixels for bpp 16, 24, 32
                         */
    
    0,                  /* size -- size of memory allocated.
                         *
                         * Leaving this zero forces subdriver to init
                         * size to psd->yres * psd->linelen.
                         */
    
    0,                  /* ncolors -- # screen colors */
    0,                  /* pixtype -- format of pixel value */
    0,                  /* flags -- device flags */
    
    (void *)VID_MEMBASE,	/* addr -- address of memory allocated.
                         *
                         * VID_MEMBASE defined in
                         * <lowlevel/vid_regs.h>
                         */
                        /* FIXME  as above. */
                        
    FB7k_open,
    FB7k_close,
    FB7k_getscreeninfo,
    NULL,		/* SetPalette subdriver */
    NULL,               /* DrawPixel subdriver */
    NULL,               /* ReadPixel subdriver */
    NULL,               /* DrawHorzLine subdriver */
    NULL,               /* DrawVertLine subdriver */
    NULL,               /* FillRect subdriver */
    gen_fonts,
    NULL,		/* DrawText */
    NULL,		/* Line */
    NULL,               /* Blit subdriver */
    NULL,               /* PreSelect*/
    NULL,               /* DrawArea subdriver*/
    NULL,               /* SetIOPermissions*/
    gen_allocatememgc,
    FB7k_mapmemgc,        /* MapMemGC*/
    gen_freememgc       /* FreeMemGC*/
};

static SUBDRIVER FB7k_subdriver = {
        FB7k_init,
        FB7k_drawpixel,
        FB7k_readpixel,
        FB7k_drawhorzline,
        FB7k_drawvertline,
        FB7k_drawfillrect,
        FB7k_blit,
        NULL,                   /* FIXME DrawArea */
        FB7k_stretchblit,
        FB7k_conditioned_line
};

/* lookup table to convert microwin modes to S1D13506 ROPs
 * use with MWROP as MWmodeto506ROP[MWROP_TO_MODE(MWROP_xxx)] */
static u_int MWmodeto506ROP[MWMODE_MAX+1] = {
	0xc,	/* MWMODE_COPY (src) */
	0x6,	/* MWMODE_XOR (src ^ dst) */
	0xe,	/* MWMODE_OR (src | dst) */
	0x8,	/* MWMODE_AND (src & dst) */
	0x0,	/* MWMODE_CLEAR (0) */
	0xf,	/* MWMODE_SETTO1 (~0) */
	0x9,	/* MWMODE_EQUIV (~(src ^ dst)) */
	0x1,	/* MWMODE_NOR (~(src | dst)) */
	0x7,	/* MWMODE_NAND (~(src & dst)) */
	0x5,	/* MWMODE_INVERT (~dst) */
	0x3,	/* MWMODE_COPYINVERTED (~src) */
	0xb,	/* MWMODE_ORINVERTED (~src | dst) */
	0x2,	/* MWMODE_ANDINVERTED (~src & dst) */
	0xd,	/* MWMODE_ORREVERSE (src | ~dst) */
	0x4,	/* MWMODE_ANDREVERSE (src & ~dst) */
	0xa	/* MWMODE_NOOP (dst) */
};

typedef enum {
	S1D13505,
	S1D13506
} vidchip_e;

typedef struct modeline_t {
	float	pclk;		/* pixel clock */
	u_short	h_res,		/* horizontal resolution */
		h_deadpixels,	/* dead horizontal pixels */
		h_nondisp,	/* horizontal nondisplay period */
		h_syncstart,	/* horizontal sync start */
		h_syncwidth,	/* horizontal sync width */
		h_bytes,	/* how many bytes in a line */
		v_res,		/* vertical resolution */
		v_nondisp,	/* vertical nondisplay period */
		v_syncstart,	/* vertical sync start */
		v_syncwidth,	/* vertical sync width */
		h_syncplus,	/* horizontal positive polarity */
		v_syncplus;	/* veritcal positive polarity */
} modeline_t;

/*
 * routines
 */

/* wait for BitBLT to finish on '506 */
static volatile void BitBLTspin(void) {
	while((VID506_REG(VID506_BITBLT_CTRL0) & 0x80));
}

/* find out what chip we have in the box */
static vidchip_e vid_detect(void) {
	u_char readreg;

	/* poke at S1D13505 */
	VID505_REG(VID505_MISC) = 0;

	/* check for S1D13505 */
	readreg = VID505_REG(VID505_REVISION_CODE);
	if (readreg == 0x0c) {
		dprintf("<vid_detect> found S1D13505\n");
		return S1D13505;
	}

	/* poke at S1D13506 */
	VID506_REG(VID506_MISC) = 0;

	/* check for S1D13506 */
	readreg = VID506_REG(VID506_REVISION_CODE);
	if (readreg == 0x11) {
		dprintf("<vid_detect> found S1D13506\n");
		return S1D13506;
	}

	/* didn't find anything */
	dprintf("<vid_detect> couldn't detect video chip\n");
	return -1;
}

/* parse a modeline into struct modeline */
static int parse_modeline(
    char *modestring,
    modeline_t *modeline,
    int bpp
) {
	char *tokenized[14];
	int loopy;

	/* set some defaults */
	modeline->h_syncplus = FALSE;
	modeline->v_syncplus = FALSE;

	/* start tokenizing */
	tokenized[0] = strtok(modestring, " \t");

	/* if we can't get the first token, forget it */
	if (tokenized[0] == NULL) {
		dprintf("<parse_modeline> couldn't tokenize modeline:\n"
		    "\t%s\n", modestring);
		return 1;
	}

	/* pixel clock */
	modeline->pclk = atof(tokenized[0]);
	
	/* check pixel clock sanity; cheap and cheezy */
	if (modeline->pclk > VID_ICLK + 0.01
	    && modeline->pclk < VID_ICLK - 0.01) {
		dprintf("<parse_modeline> pclk of %f doesn't match "
		    "hardware iclk of %f\n", modeline->pclk, VID_ICLK);
		return 1;
	}

	/* loop through and tokenize the rest of the modeline */
	for(loopy = 1; loopy < 14; loopy++) {
		tokenized[loopy] = strtok(NULL, " \t");
		if(tokenized[loopy] == NULL) break;
	}

	/* count arguments */
	if(loopy < 9) {
		dprintf("<parse_modeline> not enough tokens\n");
		return 1;
	}

	/* check for hsync and vsync tokens in positions 9 and 10, (if
	 * we even have 9 or 10 tokens...)
	 * XXX we're assuming hsync and vsync are always at positions 9
	 * XXX and 10, respectively */
	if (loopy > 9 && tokenized[9][1] == 'h'
	    && tokenized[9][0] == '+') modeline->h_syncplus = TRUE;
	if (loopy > 10 && tokenized[10][1] == 'v'
	    && tokenized[10][0] == '+') modeline->v_syncplus = TRUE;

	/* fill in structure */
	/* FIXME need so many extra sanity checks... */
	modeline->h_res 	= atoi(tokenized[1]);
	modeline->h_nondisp 	= atoi(tokenized[4]) - modeline->h_res;
	modeline->h_syncstart 	= atoi(tokenized[2]) - modeline->h_res;
	modeline->h_syncwidth 	= atoi(tokenized[3]) - atoi(tokenized[2]);
	modeline->v_res 	= atoi(tokenized[5]);
	modeline->v_nondisp	= atoi(tokenized[8]) - modeline->v_res;
	modeline->v_syncstart	= atoi(tokenized[6]) - modeline->v_res;
	modeline->v_syncwidth	= atoi(tokenized[7]) - atoi(tokenized[6]);

	/* deadpixels */
	if(modeline->h_res > 767 && bpp > 8)
	    modeline->h_deadpixels = 1024 - modeline->h_res;
	else modeline->h_deadpixels = 0;

	/* bytes per row */
	modeline->h_bytes =
	    ((modeline->h_res + modeline->h_deadpixels) * bpp) / 8;

	/* all OK here */
	return 0;
}

/* video initialization, S1D13505 version */
static int vid505_setmode(modeline_t *modeline, int bpp) {
	u_char regvalue=0;

	/* only 25.175 is supported for the old video controller */
	assert (VID_ICLK == 25.175);

	/* first the DRAM settings:
	 * 2-CAS#, EDO, refresh at ICLK/256 (98kHz).
	 * FIXME we probably won't be switching DRAM speeds anytime
	 * soon, but these will need to be changed if we do. */
	VID505_REG(VID505_MEMORY_CONFIG) = (u_char) 0x02;
	VID505_REG(VID505_PERF_ENHANCEMENT0) = (u_char) 0x24;

	/* width */
	VID505_REG(VID505_HORZ_DISP_WIDTH) = modeline->h_res / 8 - 1;

	/* horizontal non-display period */
	VID505_REG(VID505_HORZ_NONDISP_PERIOD) = modeline->h_nondisp / 8 - 1;

	/* horizontal sync start */
	VID505_REG(VID505_HRTC_START_POSITION) =
	    (modeline->h_syncstart + 2) / 8 - 1;

	/* horizontal sync width and polarity */
	VID505_REG(VID505_HRTC_PULSE_WIDTH) =
	    ((modeline->h_syncplus ? 0x80 : 0)|(modeline->h_syncwidth/8-1));

	/* height */
	VID505_REG(VID505_VERT_DISP_HEIGHT0) =
	    (modeline->v_res & 0xff) - 1;
	VID505_REG(VID505_VERT_DISP_HEIGHT1) =
	    modeline->v_res >> 8;

	/* vertical nondisplay period */
	VID505_REG(VID505_VERT_NONDISP_PERIOD) = modeline->v_nondisp - 1;

	/* vertical sync start position */
	VID505_REG(VID505_VRTC_START_POSITION) = modeline->v_syncstart - 1;

	/* vertical sync width and polarity */
	VID505_REG(VID505_VRTC_PULSE_WIDTH) =
	    (modeline->v_syncplus ? 0x80 : 0) | (modeline->v_syncwidth - 1);

	/* words per video line FIXME probably a better way to do this */
	VID505_REG(VID505_MEM_ADDR_OFFSET0) =
	    (modeline->h_res + modeline->h_deadpixels) * bpp / 16;
	VID505_REG(VID505_MEM_ADDR_OFFSET1) = 
	    ((modeline->h_res + modeline->h_deadpixels) * bpp / 16)>>8;

	/* turn on the display, and set FIFO threshold */
	switch (bpp) {
	case 16: case 15:
		VID505_REG(VID505_PERF_ENHANCEMENT1) = 0x1b;
		if (bpp == 16) regvalue = 5;
		else regvalue = 4;
		break;
	case 8: case 4: case 2: case 1:
		/* FIXME there's got to be a better technique */
		VID505_REG(VID505_PERF_ENHANCEMENT1) = 0;
		switch (bpp) {
		case 8:
			regvalue = 3;
			break;
		case 4:
			regvalue = 2;
			break;
		case 2:
			regvalue = 1;
			break;
		case 1:
			regvalue = 0;
			break;
		}
		break;
	}
	/* set bpp and turn on CRT */
	regvalue = (regvalue << 2) + 2;
	VID505_REG(VID505_DISPLAY_MODE) = regvalue;

	return 0;
}

/* video initialization, S1D13506 version */
static int vid506_setmode(modeline_t *modeline, int bpp) {
	u_char regvalue=0;

	/* disable all output devices */
	VID506_REG(VID506_DISPLAY_MODE) = 0;

	/* GPIO pins:
	 * 56	MA9 (DRAM memory address)
	 * 57	GPIO2 (ID0 from monitor)
	 * 59	GPIO1 (DSCL to/from monitor) */
	VID506_REG(VID506_GPIO_CONFIG) = 0;
	VID506_REG(VID506_GPIO_CTRL) = 0x02;

	/* clock inputs: MCLK is full-speed, use BUSCLK for
	 * MCLK; apparently you have to do some magic with bit 4 if you
	 * are toggling bit 0 */
	VID506_REG(VID506_MEM_CLOCK_CFG)|=0x10;		/* set bit 4 */
	VID506_REG(VID506_MEM_CLOCK_CFG)|=0x01;		/* set bit 0 */
	VID506_REG(VID506_MEM_CLOCK_CFG)&=~0x10;	/* clear bit 4 */

	/* set LCD PCLK to use full speed ICLK */
	VID506_REG(VID506_LCD_PCLK_CFG) = 0;

	/* set clock input for CRT, use CLKI at 1:1 ratio */
	VID506_REG(VID506_CRTTV_PCLK_CFG) = 0;

	/* divide down media plug clock, since that's what the example
	 * code does.  */
	VID506_REG(VID506_MPCLK_CFG) = 0x03;

	/* wait states: BCLK is ~24.117MHz, MCLK is at least 25.175MHz,
	 * so no need for wait states. */
	VID506_REG(VID506_CPU2MEM_WAIT_SEL) = 0;

	/* DRAM configuration:
	 * 2-CAS#, EDO */
	VID506_REG(VID506_MEM_CFG) = 0;

	/* need refresh of at least 64kHz for the current 50ns EDO-DRAM.
	 * calculate it. */
	{
		char refresh = 7;
		int iclk = (VID_ICLK * 1000.0);
		int rate = iclk / (1<<(refresh + 6));

		while (rate < 64 && refresh > 0)
		{
			refresh--;
			rate = iclk / (1<<(refresh + 6));
		}

		assert(refresh >= 0);
		if (refresh < 0) return 1;

		dprintf("VID506_DRAM_REFRESH: 0x%02x, (%dkHZ)\n",
			refresh, rate);

		VID506_REG(VID506_DRAM_REFRESH) = 0x02;
	}

	/* self-refresh when in power save mode; timing per Epson
	 * reccomended specs.  Epson doesn't tell us how to
	 * calculate these, so we're effectively left with a lookup
	 * table */
	{
		unsigned char ctrl0, ctrl1;
		int iclk = (VID_ICLK * 1000.0);

		switch(iclk)
		{
			case 25175:
			case 31500:
			case 32000:
			case 33000:
				ctrl0 = 0x12;
				ctrl1 = 0x02;
				break;
			case 40000:
				ctrl0 = 0x01;
				ctrl1 = 0x01;
				break;
			default:
				assert(0);
				return 1;
		}

		VID506_REG(VID506_DRAM_TIMINGS_CTRL0) = ctrl0;
		VID506_REG(VID506_DRAM_TIMINGS_CTRL1) = ctrl1;
	}

	/* XXX LCD panel foo (I shouldn't need to be here */
	VID506_REG(VID506_PANEL_TYPE) = 0x25;
	VID506_REG(VID506_MOD_RATE) = 0x00;
	VID506_REG(VID506_LCD_HDP) = 0x4f;
	VID506_REG(VID506_LCD_HNDP) = 0x13;
	VID506_REG(VID506_TFT_FPLINE_START) = 0x01;
	VID506_REG(VID506_TFT_FPLINE_PULSE) = 0x0b;
	VID506_REG(VID506_LCD_VDP0) = 0xdf;
	VID506_REG(VID506_LCD_VDP1) = 0x01;
	VID506_REG(VID506_LCD_VNDP) = 0x2c;
	VID506_REG(VID506_TFT_FPFRAME_START) = 0x0b;
	VID506_REG(VID506_TFT_FPFRAME_PULSE) = 0x01;

	/* XXX LCD controller foo */
	VID506_REG(VID506_LCD_DISPLAY_MODE) = 0x83;
	VID506_REG(VID506_LCD_MISC) = 0x00;
	VID506_REG(VID506_LCD_START_ADDR0) = 0;
	VID506_REG(VID506_LCD_START_ADDR1) = 0;
	VID506_REG(VID506_LCD_START_ADDR2) = 0;
	VID506_REG(VID506_LCD_MEM_ADDR_OFFSET0) = 0;
	VID506_REG(VID506_LCD_MEM_ADDR_OFFSET1) = 0;
	VID506_REG(VID506_LCD_PIXEL_PANNING) = 0;
	VID506_REG(VID506_LCD_FIFO_HIGH_THRESHOLD) = 0;
	VID506_REG(VID506_LCD_FIFO_LOW_THRESHOLD) = 0;

	/* horizontal width */
	VID506_REG(VID506_CRTTV_HDP) = modeline->h_res / 8 - 1;

	/* horizontal non-display width */
	VID506_REG(VID506_CRTTV_HNDP) = modeline->h_nondisp / 8 - 1;

	/* horizontal sync start */
	VID506_REG(VID506_CRTTV_HRTC_START) =
	    (modeline->h_syncstart + 2) / 8 - 1;

	/* horizontal sync width and polarity later */
	VID506_REG(VID506_CRT_HRTC_PULSE) =
	    (modeline->h_syncplus?0x80:0)|(modeline->h_syncwidth / 8 - 1);

	/* height */
	VID506_REG(VID506_CRTTV_VDP0) = (modeline->v_res & 0xff) - 1;
	VID506_REG(VID506_CRTTV_VDP1) = modeline->v_res >> 8;

	/* vertical nondisplay */
	VID506_REG(VID506_CRTTV_VNDP) = modeline->v_nondisp - 1;

	/* vertical sync width */
	VID506_REG(VID506_CRTTV_VRTC_START) = modeline->v_syncstart - 1;

	/* vertical retrace pulse width register & polarity? */
	VID506_REG(VID506_CRT_VRTC_PULSE) =
	  (modeline->v_syncplus?0x80:0)|(modeline->v_syncwidth - 1);

	/* CRT/TV output control: we use 4.6mA IREF */
	VID506_REG(VID506_TV_OUTPUT_CTRL) = 0x08;

	/* start the refresh pipeline and set bpp for the CRT */
	switch (bpp) {
	case 16: regvalue = 5; break;
	case 15: regvalue = 4; break;
	case 8: regvalue = 3; break;
	case 4: regvalue = 2; break;
	default: return 1; break;
	}
	VID506_REG(VID506_CRTTV_DISPLAY_MODE) = regvalue;

	/* what we show on the screen starts at DRAM address 0 */
	VID506_REG(VID506_CRTTV_START_ADDR0) = 0;
	VID506_REG(VID506_CRTTV_START_ADDR1) = 0;
	VID506_REG(VID506_CRTTV_START_ADDR2) = 0;

	/* words per video line; no virtual viewport == no offset */
	VID506_REG(VID506_CRTTV_MEM_ADDR_OFFSET0) =
	    (modeline->h_res + modeline->h_deadpixels) * bpp / 16;
	VID506_REG(VID506_CRTTV_MEM_ADDR_OFFSET1) =
	    ((modeline->h_res + modeline->h_deadpixels) * bpp / 16)>>8;

	/* no panning */
	VID506_REG(VID506_CRTTV_PIXEL_PANNING) = 0;

	/* FIFO fiddles, set to auto-tune.  not critical for single
	 * display applications, apparently. */
	VID506_REG(VID506_CRTTV_FIFO_HIGH_THRESHOLD) = 0;
	VID506_REG(VID506_CRTTV_FIFO_LOW_THRESHOLD) = 0;

	/* XXX LCD INK */
	VID506_REG(VID506_LCD_INK_CURS_CTRL) = 0;
	VID506_REG(VID506_LCD_INK_CURS_START_ADDR) = 0;
	VID506_REG(VID506_LCD_CURSOR_X_POS0) = 0;
	VID506_REG(VID506_LCD_CURSOR_X_POS1) = 0;
	VID506_REG(VID506_LCD_CURSOR_Y_POS0) = 0;
	VID506_REG(VID506_LCD_CURSOR_Y_POS1) = 0;
	VID506_REG(VID506_LCD_INK_CURS_BLUE0) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_GREEN0) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_RED0) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_BLUE1) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_GREEN1) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_RED1) = 0x00;
	VID506_REG(VID506_LCD_INK_CURS_FIFO) = 0x00;

	/* XXX CRT/TV INK */
	VID506_REG(VID506_CRTTV_INK_CURS_CTRL) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_START_ADDR) = 0x00;
	VID506_REG(VID506_CRTTV_CURSOR_X_POS0) = 0x00;
	VID506_REG(VID506_CRTTV_CURSOR_X_POS1) = 0x00;
	VID506_REG(VID506_CRTTV_CURSOR_Y_POS0) = 0x00;
	VID506_REG(VID506_CRTTV_CURSOR_Y_POS1) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_BLUE0) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_GREEN0) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_RED0) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_BLUE1) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_GREEN1) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_RED1) = 0x00;
	VID506_REG(VID506_CRTTV_INK_CURS_FIFO) = 0x00;

	/* XXX BitBLT */
	VID506_REG(VID506_BITBLT_CTRL0) = 0x00;
	VID506_REG(VID506_BITBLT_CTRL1) = 0x00;
	VID506_REG(VID506_BITBLT_ROP_CODE) = 0x00;
	VID506_REG(VID506_BITBLT_OPERATION) = 0x00;
	VID506_REG(VID506_BITBLT_SRC_START_ADDR0) = 0x00;
	VID506_REG(VID506_BITBLT_SRC_START_ADDR1) = 0x00;
	VID506_REG(VID506_BITBLT_SRC_START_ADDR2) = 0x00;
	VID506_REG(VID506_BITBLT_DEST_START_ADDR0) = 0x00;
	VID506_REG(VID506_BITBLT_DEST_START_ADDR1) = 0x00;
	VID506_REG(VID506_BITBLT_DEST_START_ADDR2) = 0x00;
	VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) = 0x00;
	VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) = 0x00;
	VID506_REG(VID506_BITBLT_WIDTH0) = 0x00;
	VID506_REG(VID506_BITBLT_WIDTH1) = 0x00;
	VID506_REG(VID506_BITBLT_HEIGHT0) = 0x00;
	VID506_REG(VID506_BITBLT_HEIGHT1) = 0x00;
	VID506_REG(VID506_BITBLT_BACKGND_COLOR0) = 0x00;
	VID506_REG(VID506_BITBLT_BACKGND_COLOR1) = 0x00;
	VID506_REG(VID506_BITBLT_FOREGND_COLOR0) = 0x00;
	VID506_REG(VID506_BITBLT_FOREGND_COLOR1) = 0x00;

	/* XXX LUT */
	VID506_REG(VID506_LUT_MODE) = 0x02;
	VID506_REG(VID506_LUT_ADDR) = 0;
	VID506_REG(VID506_LUT_DATA) = 0;
#if 1
	/* XXX test to see if we can make black black */
	VID506_REG(VID506_LUT_DATA) = 0;
	VID506_REG(VID506_LUT_DATA) = 0;
#endif

	/* disable power saving */
	VID506_REG(VID506_PWR_SAVE_CFG) = 0;
	VID506_REG(VID506_PWR_SAVE_STATUS) = 0;

	/* disable the watchdog */
	VID506_REG(VID506_CPU2MEM_WATCHDOG) = 0;

	/* fire it up */
	VID506_REG(VID506_DISPLAY_MODE) = 0x02;

	/* all done? */
	return 0;
}

/* Calc linelen and mmap size, return 0 on fail*/
static int
FB7k_init(PSD psd)
{
	if (!psd->size)
		psd->size = psd->yres * psd->linelen;
	/* linelen in bytes for bpp 1, 2, 4, 8 so no change*/
	return 1;
}

/* Set pixel at x, y, to pixelval c*/
static void
FB7k_drawpixel(PSD psd, MWCOORD x, MWCOORD y, MWPIXELVAL c)
{
	ADDR8	addr = psd->addr;

	assert (addr != 0);
	assert (x >= 0 && x < psd->xres);
	assert (y >= 0 && y < psd->yres);
	assert (c < psd->ncolors);

	DRAWON;
	if(gr_mode == MWMODE_COPY)
		addr[x + y * psd->linelen] = c;
	else
		applyOp(gr_mode, c, &addr[ x + y * psd->linelen], ADDR8);
	DRAWOFF;
}

/* Read pixel at x, y*/
static MWPIXELVAL
FB7k_readpixel(PSD psd, MWCOORD x, MWCOORD y)
{
	ADDR8	addr = psd->addr;

	assert (addr != 0);
	assert (x >= 0 && x < psd->xres);
	assert (y >= 0 && y < psd->yres);

	return addr[x + y * psd->linelen];
}

/* Draw horizontal line from x1,y to x2,y including final point
 *
 * must handle dashlists and stipples */
static void
FB7k_drawhorzline(PSD psd, MWCOORD x1, MWCOORD x2, MWCOORD y, MWPIXELVAL c)
{
	ADDR8	addr = psd->addr;
	extern unsigned long	gr_dashmask;
	extern unsigned long	gr_dashcount;
	extern char		gr_dashindex;
	extern int		gr_linestyle;
	extern MWSTIPPLE	gr_stipple;

        /* sanity checks */
	assert (addr != 0);
	assert (x1 >= 0 && x1 < psd->xres);
	assert (x2 >= 0 && x2 < psd->xres);
	assert (x2 >= x1);
	assert (y >= 0 && y < psd->yres);
	assert (c < psd->ncolors);
	assert (gr_stipple.width == 0 && gr_stipple.height == 0);

        /* start drawing */
	DRAWON;

        /* set initial start point */
	addr += x1 + y * psd->linelen;

        /* are we using some weird mode, like XOR? */
	if(gr_mode != MWMODE_COPY) {
                if (gr_dashmask == 0 || gr_linestyle == MWLINE_SOLID) {
                        while(x1++ <= x2) {
                                applyOp(gr_mode, c, addr, ADDR8);
                                ++addr;
                        }
                } else {
                        while(x1++ <= x2) {
                                if (gr_dashmask & (1 << gr_dashindex)) 
                                    applyOp(gr_mode, c, addr, ADDR8);
                                addr++;
                                gr_dashindex++;
                                gr_dashindex%=gr_dashcount;
                        }
                }
	} else if (gr_dashmask == 0 || gr_linestyle == MWLINE_SOLID) {
                /* XXX precludes more than 32768 pixels ;) */
                short   pixelstodraw = x2 - x1 + 1;

                /* if we're unaligned, write pixels until we are.  */
                if((unsigned long)addr & 0x1) {
                        *addr++ = c;
                        pixelstodraw--;
                }

                /* do we have a lot of pixels to draw? */
                if (pixelstodraw >= sizeof(unsigned long) * 2) {
                        /* we can write using this many 32-bit
                         * operations */
                        short   longstodraw =
                            pixelstodraw/sizeof(unsigned long);
                        /* assemble our paintbrush */
                        unsigned long   brush32;

                        brush32 = c&0xff;
                        brush32|= brush32<<8;
                        brush32|= brush32<<16;

                        /* paint */
			while(longstodraw--) {
				*(unsigned long *)addr = brush32;
				addr+=sizeof(unsigned long);
			}

                        /* wash down our brush */
                        pixelstodraw%=sizeof(unsigned long);
                }

                /* trim work */
                while(pixelstodraw--) *addr++ = c;
        } else {
		/* alternate method for testing... just use byte
		 * manipulations */
		short	pixelstodraw = x2 - x1 + 1;

                while(pixelstodraw--) {
			if (gr_dashmask & (1 << gr_dashindex)) *addr = c;
			addr++;
			gr_dashindex++;
			gr_dashindex%=gr_dashcount;
		}
	}
        /* done */
	DRAWOFF;
}

/* Draw a vertical line from x,y1 to x,y2 including final point*/
static void
FB7k_drawvertline(PSD psd, MWCOORD x, MWCOORD y1, MWCOORD y2, MWPIXELVAL c)
{
	ADDR8	addr = psd->addr;
	int	linelen = psd->linelen;
	extern unsigned long	gr_dashmask;
	extern unsigned long	gr_dashcount;
	extern char		gr_dashindex;
	extern int		gr_linestyle;

	assert (addr != 0);
	assert (x >= 0 && x < psd->xres);
	assert (y1 >= 0 && y1 < psd->yres);
	assert (y2 >= 0 && y2 < psd->yres);
	assert (y2 >= y1);
	assert (c < psd->ncolors);

	DRAWON;
	addr += x + y1 * linelen;
	if(gr_mode == MWMODE_COPY) {
                short pixelstodraw = y2 - y1 + 1;
		if (gr_dashmask == 0 || gr_linestyle == MWLINE_SOLID) {
			while(pixelstodraw--) {
				*addr = c;
				addr+=linelen;
			}
		} else {
			while(pixelstodraw--) {
				if (gr_dashmask & (1 << gr_dashindex))
				    *addr = c;
				addr+=linelen;
				gr_dashindex++;
				gr_dashindex%=gr_dashcount;
			}
		}
	} else {
		while(y1++ <= y2) {
			applyOp(gr_mode, c, addr, ADDR8);
			addr+=linelen;
		}
	}
	DRAWOFF;
}

/* filled rectangle */
static void FB7k_drawfillrect(PSD psd,MWCOORD x1, MWCOORD y1, MWCOORD x2,
	MWCOORD y2, MWPIXELVAL c)
{
        /* sanity checks */
	assert (psd->addr != 0);
	assert (x1 >= 0 && x1 < psd->xres);
	assert (x2 >= 0 && x2 < psd->xres);
	assert (x2 >= x1);
	assert (y1 >= 0 && y1 < psd->yres);
	assert (y2 >= 0 && y2 < psd->yres);
	assert (c < psd->ncolors);

	DRAWON;

        /* set initial start point */

	/* standard drawing mode? */
	if(gr_mode != MWMODE_COPY) {
		/* the slow safe method */
		while(y1 <= y2) {
			MWCOORD curx1 = x1;
			ADDR8 addr = psd->addr + x1 + y1 * psd->linelen;
			while(curx1++ <= x2) {
				applyOp(gr_mode, c, addr, ADDR8);
				++addr;
			}
			y1++;
		}
	} else {
		/* speed speed speed */
		ADDR8		linestartaddr =
		    psd->addr + x1 + y1 * psd->linelen,
		    		addr;
		short		linepixelstodraw = x2 - x1 + 1;

		/* is it worth the trouble? */
		if (linepixelstodraw > sizeof(unsigned long) * 2) {
			short		beginpixels = 0, endpixels = 0,
					middlelongs;
			unsigned long	brush32;

			/* make a paintbrush */
			brush32 = c&0xff;
			brush32|= brush32<<8;
			brush32|= brush32<<16;

			/* calculate starting byte-sized pixels */
			if ((unsigned long)linestartaddr & 0x1) {
				beginpixels = 1;
				linepixelstodraw--;
			}

			/* calculate ending byte-sized pixels */
			endpixels =
			    linepixelstodraw%sizeof(unsigned long);

			/* calculate longs for the middle */
			middlelongs =
			    linepixelstodraw / sizeof(unsigned long);

			/* since we are guaranteed even-sized line
			 * lengths, this is the same every iteration */
			while(y1++ <= y2) {
				short loopy;
				addr = linestartaddr;

				/* begin */
				if(beginpixels) *addr++ = c;

				/* middle */
				loopy = middlelongs;
				while(loopy--) {
					*(unsigned long *)addr = brush32;
					addr+=sizeof(unsigned long);
				}

				/* end */
				loopy = endpixels;
				while(loopy--) *addr++ = c;

				/* lather, rinse, and repeat */
				linestartaddr += psd->linelen;
			}

		} else {
			/* we have tiny rows */
			short	linepixelstodraw = x2 - x1 + 1;

			while(y1++ <= y2) {
				short loopy;

				addr = linestartaddr;
				loopy = linepixelstodraw;

				while(loopy--) *addr++ = c;

				linestartaddr += psd->linelen;
			}
		}
	}

	DRAWOFF;
}

/* accelerated filled rectangle
 * may not necessarily be called only for objects in video memory! */
void FB7kBitBLT_drawfillrect(PSD psd, MWCOORD x1, MWCOORD y1,
	MWCOORD x2, MWCOORD y2, MWPIXELVAL c)
{
	u_int	screenoffset, startoffset;
	u_short	width, height, stride;

        /* sanity checks */
	assert (psd->addr != 0);
	assert (x1 >= 0 && x1 < psd->xres);
	assert (x2 >= 0 && x2 < psd->xres);
	assert (x2 >= x1);
	assert (y1 >= 0 && y1 < psd->yres);
	assert (y2 >= 0 && y2 < psd->yres);
	assert (y2 >= y1);
	assert (c < psd->ncolors);
	assert ((psd->linelen%2) == 0);
	assert (((x2 - x1) & 0xfc00) == 0);
#if 0
	assert (((y2 - y1) & 0xfc00) == 0);
#endif

	/* if this isn't video memory, can't use the bitblit engine */
	if ((u_int)psd->addr < VID_MEMBASE
	    || (u_int)psd->addr > (VID_MEMBASE + VID_MEMSIZE)) {
		FB7k_drawfillrect(psd, x1, y1, x2, y2, c);
		return;
	}

	/* this really is a filled rectangle routine only... can't
	 * handle any other operations right now */
	if (gr_mode != MWMODE_COPY) {
		FB7k_drawfillrect(psd, x1, y1, x2, y2, c);
		return;
	}

	/* if our height is > 10 bits, a bunch of 10 bits worth, then
	 * continue on with the leftovers */
	while (y2 - y1 > 0x3ff) {
		FB7kBitBLT_drawfillrect(psd, x1, y1, x2, y1+0x3ff, c);
		y1+=0x400;
	}

	ddprintf("<FB7kBitBLT_drawfillrect> %d,%d-%d,%d (%p)\n",
	    x1, y1, x2, y2, psd->addr);

	/* if we are asked to fill the full width with black, and the
	 * width has been rounded up, take the chance to set the last
	 * column to the transparent color, 0xff */
	if (x1 == 0 && c == 0 && psd->xres < psd->linelen &&
	    x2 == (psd->xres - 1)
	) {

		/* calculate what we're going to give the
		 * controller */
		startoffset = (u_int)psd->addr - VID_MEMBASE + psd->xres
		    + y1 * psd->linelen;
		width = psd->linelen - psd->xres - 1;
		assert ((width & 0xfc00) == 0);
		height = y2 - y1;
		stride = psd->linelen / 2;

		ddprintf("<FB7kBitBLT_drawfillrect> * 0x%x, %dx%d, +%d\n",
		    startoffset, width, height, stride);

		DRAWON;

		/* starting offset is the little off-screen area */
		VID506_REG(VID506_BITBLT_DEST_START_ADDR0) =
		    startoffset & 0xff;
		startoffset>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR1) =
		    startoffset & 0xff;
		startoffset>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR2) =
		    startoffset & 0xff;

		/* width is pixels - 1 */
		VID506_REG(VID506_BITBLT_WIDTH0) = width & 0xff;
		width>>=8;
		VID506_REG(VID506_BITBLT_WIDTH1) = width & 0xff;
		VID506_REG(VID506_BITBLT_HEIGHT0) = height & 0xff;
		height>>=8;
		VID506_REG(VID506_BITBLT_HEIGHT1) = height & 0xff;

		/* set the colors */
		VID506_REG(VID506_BITBLT_FOREGND_COLOR0) = 0xff;

		/* solid fill */
		VID506_REG(VID506_BITBLT_OPERATION) = 0x0c;

		/* for completeness, set 8bpp */
		VID506_REG(VID506_BITBLT_CTRL1) = 0;

		/* set screen stride */
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) =
		    stride & 0xff;
		stride>>=8;
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) =
		    stride & 0xff;

		/* definitely a rectangular region */
		VID506_REG(VID506_BITBLT_CTRL0) = 0x80;

		BitBLTspin();

		DRAWOFF;
	}

	DRAWON;

	/* calculate what we need to stuff into the registers */
	screenoffset = (u_int)psd->addr - VID_MEMBASE;
	startoffset = screenoffset + x1 + y1 * psd->linelen;
	width = x2 - x1;
	height = y2 - y1;
	stride = psd->linelen / 2;

	ddprintf("<FB7kBitBLT_drawfillrect> 0x%x, %dx%d, +%d\n",
	    startoffset, width, height, stride);

	/* stuff it into the proper BitBLT registers. */
	VID506_REG(VID506_BITBLT_DEST_START_ADDR0) = startoffset & 0xff;
	startoffset>>=8;
	VID506_REG(VID506_BITBLT_DEST_START_ADDR1) = startoffset & 0xff;
	startoffset>>=8;
	VID506_REG(VID506_BITBLT_DEST_START_ADDR2) = startoffset & 0xff;

	/* set the width and height */
	VID506_REG(VID506_BITBLT_WIDTH0) = width & 0xff;
	width>>=8;
	VID506_REG(VID506_BITBLT_WIDTH1) = width & 0xff;
	VID506_REG(VID506_BITBLT_HEIGHT0) = height & 0xff;
	height>>=8;
	VID506_REG(VID506_BITBLT_HEIGHT1) = height & 0xff;

	/* set the colors */
	VID506_REG(VID506_BITBLT_FOREGND_COLOR0) = c;

	/* solid fill */
	VID506_REG(VID506_BITBLT_OPERATION) = 0x0c;

	/* for completeness, set 8bpp */
	VID506_REG(VID506_BITBLT_CTRL1) = 0;

	/* set screen stride */
	VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) = stride & 0xff;
	stride>>=8;
	VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) = stride & 0xff;

	/* offscreen regions are rectangular too, just that the stride
	 * can change. */
	VID506_REG(VID506_BITBLT_CTRL0) = 0x80;

	/* wait for bitblit to finish... for onesie twosie bitblit uses,
	 * we are forced to spin before continuing so other people don't
	 * hit video memory and disturb the bitblt engine; ideally we
	 * could check at the beginning which would allow us to trundle
	 * on while the bitblt does its thing.  */
	BitBLTspin();

	DRAWOFF;

	ddoff;
}

/* srccopy bitblt*/
static void
FB7k_blit(PSD dstpsd, MWCOORD dstx, MWCOORD dsty, MWCOORD w, MWCOORD h,
	PSD srcpsd, MWCOORD srcx, MWCOORD srcy, long op)
{
	ADDR8	dst;
	ADDR8	src;
	int	dlinelen = dstpsd->linelen;
	int	slinelen = srcpsd->linelen;

	assert (dstpsd->addr != 0);
	assert (dstx >= 0 && dstx < dstpsd->xres);
	assert (dsty >= 0 && dsty < dstpsd->yres);
	assert (w > 0);
	assert (h > 0);
	assert (srcpsd->addr != 0);
	assert (srcx >= 0 && srcx < srcpsd->xres);
	assert (srcy >= 0 && srcy < srcpsd->yres);
	assert (dstx+w <= dstpsd->xres);
	assert (dsty+h <= dstpsd->yres);
	assert (srcx+w <= srcpsd->xres);
	assert (srcy+h <= srcpsd->yres);

	DRAWON;
	dst = dstpsd->addr + dstx + dsty * dlinelen;
	src = srcpsd->addr + srcx + srcy * slinelen;

	if (op == MWROP_COPY) {
		/* copy from bottom up if dst in src rectangle*/
		/* memmove is used to handle x case*/
		if (srcy < dsty) {
			src += (h-1) * slinelen;
			dst += (h-1) * dlinelen;
			slinelen *= -1;
			dlinelen *= -1;
		}
                if (dstpsd == srcpsd) {
                        /* copying on the same screen device, so we have
                         * to do overlap checks for each line copied and
                         * act accordingly... */
                        while(--h >= 0) {
                                /* check for overlap */
                                if (src < dst && dst < src + w) {

                                        /* copy backwards by bytes
                                         * XXX this could be optimized more */
                                        src += w;
                                        dst += w;
                                        {
                                                short loopy = w;
                                                while(loopy--) *--dst = *--src;
                                        }
                                } else if (
                                    w > 4
                                    && ((unsigned long)src%2) == 0
                                    && ((unsigned long)dst%2) == 0
                                ) {
                                        /* no overlap, more than 4 bytes, and
                                         * both word-aligned */
                                        short   longstodraw =
                                            w / sizeof(unsigned long),
                                                leftovers =
                                            w % sizeof(unsigned long);
                                        ADDR8   rsrc = src, wdst = dst;

					while(longstodraw--) {
						*(unsigned long *)wdst =
						    *(unsigned long *)rsrc;
						wdst+=sizeof(unsigned long);
						rsrc+=sizeof(unsigned long);
					}

                                        /* byte leftovers */
                                        while(leftovers--) *wdst++ = *rsrc++;
                                } else {
                                        /* no overlap, less than 4 bytes */
                                        short loopy = w;
                                        ADDR8 rsrc = src, wdst = dst;

                                        while(loopy--) *wdst++ = *rsrc++;
                                }

                                /* next go-round */
                                dst += dlinelen;
                                src += slinelen;
                        }
                } else {
                        /* check to see if per-line optimization is even
                         * worth it */
                        if (w < 4) while(--h >= 0) {
                                /* it's not, just copy the bytes */
                                short loopy = w;
                                ADDR8 rsrc = src, wdst = dst;

                                while(loopy--) *wdst++ = *rsrc++;

                                dst += dlinelen;
                                src += slinelen;
                        } else while (--h >= 0) {
                                /* we cannot make alignment assumptions
                                 * outside the loop, as slinelen and
                                 * dlinelen may or may not be
                                 * alignment-preserving */

                                /* can we use long operations? */
                                if (((unsigned long)src%2) == 0
                                    && ((unsigned long)dst%2) == 0) {
                                        /* that's a yes */
                                        short   longstodraw =
                                            w / sizeof(unsigned long),
                                                leftovers =
                                            w % sizeof(unsigned long);
                                        ADDR8   rsrc = src, wdst = dst;

                                        while(longstodraw--) {
						*(unsigned long *)wdst =
						    *(unsigned long *)rsrc;
						wdst+=sizeof(unsigned long);
						rsrc+=sizeof(unsigned long);
                                        }

                                        /* clean up byte leftovers */
                                        while(leftovers--)
                                            *wdst++ = *rsrc++;
                                } else {
                                        /* no overlap, bytewise forward
                                         * copy */
                                        short loopy = w;
                                        ADDR8 rsrc = src, wdst = dst;

                                        while(loopy--) *wdst++ = *rsrc++;
                                }

                                /* prepare for next iteration */
                                dst += dlinelen;
                                src += slinelen;
                        }
                }
	} else {
                /* slow slow but safe... */
		while (--h >= 0) {
			int i;
			for (i=0; i<w; i++) {
				applyOp(MWROP_TO_MODE(op), *src, dst, ADDR8);
				++src;
				++dst;
			}
			dst += dlinelen - w;
			src += slinelen - w;
		}
	}
	DRAWOFF;
}

/* BitBLT version of blit */
static void
FB7kBitBLT_blit(PSD dstpsd, MWCOORD dstx, MWCOORD dsty, MWCOORD w, MWCOORD h,
	PSD srcpsd, MWCOORD srcx, MWCOORD srcy, long op)
{
	u_int	dstbase, srcbase, dlinelen, slinelen, stride=0;
	u_char	blitcommand = 0x80, blitoperation = 0x02;
	u_int	dst, dstend, src, srcend, width, height;

	assert (dstpsd->addr != 0);
	assert ((u_int)dstpsd->addr >= VID_MEMBASE);
	assert ((u_int)dstpsd->addr <  VID_MEMBASE+VID_MEMSIZE);
	assert (dstx >= 0 && dstx < dstpsd->xres);
	assert (dsty >= 0 && dsty < dstpsd->yres);
	assert (w > 0);
	assert (h > 0);
	assert (srcpsd->addr != 0);
	assert ((u_int)srcpsd->addr >= VID_MEMBASE);
	assert ((u_int)srcpsd->addr <  VID_MEMBASE+VID_MEMSIZE);
	assert (srcx >= 0 && srcx < srcpsd->xres);
	assert (srcy >= 0 && srcy < srcpsd->yres);
	assert (dstx+w <= dstpsd->xres);
	assert (dsty+h <= dstpsd->yres);
	assert (srcx+w <= srcpsd->xres);
	assert (srcy+h <= srcpsd->yres);
	assert ((w & 0xfc00) == 0);
	assert ((h & 0xfc00) == 0);

	/* "housekeeping" variables to increase readability */
	dstbase = (u_int)dstpsd->addr - VID_MEMBASE;
	srcbase = (u_int)srcpsd->addr - VID_MEMBASE;
	dlinelen = dstpsd->linelen;
	slinelen = srcpsd->linelen;

#if 0
	/* XXX testing */
	if (w == slinelen) {
		FB7k_blit(
		    dstpsd, dstx, dsty, w, h, srcpsd, srcx, srcy, op
		);
		return;
	}
#endif
#if 0
	/* if linelens aren't the same then we have some linear regions */
	if (slinelen != dlinelen) {
		/* linear source */
		if (w == slinelen) blitcommand|=0x01;
		/* linear destination */
		if (w == dlinelen) blitcommand|=0x02;
	}

	/* quick sanity check */
	assert((blitcommand&0x3) != 0 || dlinelen%2 == 0);

	/* assume we're striding by the destination line length */
	stride = dlinelen>>1;
#else

	/* a hack of sorts... if we are copying the whole width of an
	 * odd-pixeled region, we can cheat and pretend we're copying a
	 * linear region by using a transparent blit.  this only works
	 * on the source side, which is the most common case for Frye. */
	if ((w+1) == slinelen) {
		/* expand our width */
		w++;

		/* if we're drawing to the screen, it needs to be
		 * transparent, otherwise just straight copy. */
		if (dstbase == 0) {
			blitoperation = 0x05;

			DRAWON;
			VID506_REG(VID506_BITBLT_BACKGND_COLOR0) = 0xff;
			VID506_REG(VID506_BITBLT_BACKGND_COLOR1) = 0x00;
			DRAWOFF;
		}
	}

	/* in order to use the region bitblit, both the source and the
	 * destination have to be rectangular with the same linelength
	 * or one of them has to be linear */
	if (slinelen != dlinelen
	    && w != slinelen
	    && w != dlinelen) {
		FB7k_blit(
		    dstpsd, dstx, dsty, w, h, srcpsd, srcx, srcy, op
		);
		ddoff;
		return;
	}

	ddprintf("<FB7kBitBLT_blit> %d,%d@%d,%d (%p) -> %d,%d (%p) op %d\n",
	    w, h, srcx, srcy, srcpsd->addr, dstx, dsty, dstpsd->addr, op);

	/* check for linear regions, adjust blitcommand accordingly */
	if (w == slinelen) blitcommand|=0x01;
	if (w == dlinelen) blitcommand|=0x02;

	/* set the stride depending on which region is linear */
	switch (blitcommand&0x3) {
	case 0:	/* source and dest are rectangular */
		assert(slinelen == dlinelen);
		stride = dlinelen>>1;
		break;
	case 1:	/* source is linear, destination is rectangular */
		assert(dlinelen%2 == 0);
		stride = dlinelen>>1;
		break;
	case 2: /* source is rectangular, destination is linear */
		assert(slinelen%2 == 0);
		stride = slinelen>>1;
		break;
	case 3:	/* source and destination are linear */
		/* FALLTHROUGH */
	default:	/* impossible */
		break;
	}
#endif
	/* calculate some common variables */
	src = srcbase + srcx + srcy * slinelen;
	dst = dstbase + dstx + dsty * dlinelen;
	srcend = src + w + h * slinelen;
	dstend = dst + w + h * slinelen;

	/* these get stuffed into registers */
	width = w - 1;
	height = h - 1;

	/* see if we can copy forwards */
	if (dstpsd != srcpsd || srcend <= dst) {

		/* yay! start the blit */
		DRAWON;

		/* copy from here... */
		VID506_REG(VID506_BITBLT_SRC_START_ADDR0) = src & 0xff;
		src>>=8;
		VID506_REG(VID506_BITBLT_SRC_START_ADDR1) = src & 0xff;
		src>>=8;
		VID506_REG(VID506_BITBLT_SRC_START_ADDR2) = src & 0xff;

		/* ...to here */
		VID506_REG(VID506_BITBLT_DEST_START_ADDR0) = dst & 0xff;
		dst>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR1) = dst & 0xff;
		dst>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR2) = dst & 0xff;

		/* set width and height */
		VID506_REG(VID506_BITBLT_WIDTH0) = width & 0xff;
		width>>=8;
		VID506_REG(VID506_BITBLT_WIDTH1) = width & 0xff;
		VID506_REG(VID506_BITBLT_HEIGHT0) = height & 0xff;
		height>>=8;
		VID506_REG(VID506_BITBLT_HEIGHT1) = height & 0xff;

		/* positive-going blit */
		VID506_REG(VID506_BITBLT_OPERATION) = blitoperation;

		/* use this operation */
		VID506_REG(VID506_BITBLT_ROP_CODE) =
		    MWmodeto506ROP[MWROP_TO_MODE(op)];

		/* we are 8bpp */
		VID506_REG(VID506_BITBLT_CTRL1) = 0;

		/* screen stride, already checked to be the same between
		 * source and destination */
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) = stride & 0xff;
		stride>>=8;
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) = stride & 0xff;

		/* here goes nothin' */
		VID506_REG(VID506_BITBLT_CTRL0) = blitcommand;
	} else {
		/* possibly overlapping, so go backwards */

		DRAWON;

		/* copy from here */
		VID506_REG(VID506_BITBLT_SRC_START_ADDR0) = srcend & 0xff;
		srcend>>=8;
		VID506_REG(VID506_BITBLT_SRC_START_ADDR1) = srcend & 0xff;
		srcend>>=8;
		VID506_REG(VID506_BITBLT_SRC_START_ADDR2) = srcend & 0xff;

		/* to here */
		VID506_REG(VID506_BITBLT_DEST_START_ADDR0) = dstend & 0xff;
		dstend>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR1) = dstend & 0xff;
		dstend>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR2) = dstend & 0xff;

		/* set width and height */
		VID506_REG(VID506_BITBLT_WIDTH0) = width & 0xff;
		width>>=8;
		VID506_REG(VID506_BITBLT_WIDTH1) = width & 0xff;
		VID506_REG(VID506_BITBLT_HEIGHT0) = height & 0xff;
		height>>=8;
		VID506_REG(VID506_BITBLT_HEIGHT1) = height & 0xff;

		/* negative-going blit */
		VID506_REG(VID506_BITBLT_OPERATION) = 0x03;

		/* use this operation */
		VID506_REG(VID506_BITBLT_ROP_CODE) =
		    MWmodeto506ROP[MWROP_TO_MODE(op)];

		/* 8bpp */
		VID506_REG(VID506_BITBLT_CTRL1) = 0;

		/* same device; same stride */
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) = stride & 0xff;
		stride>>=8;
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) = stride & 0xff;

		/* let 'er rip */
		VID506_REG(VID506_BITBLT_CTRL0) = blitcommand;
	}

	/* wait for bitblt to finish before continuing on */
	BitBLTspin();
	
	ddoff;

	DRAWOFF;
}

/* srccopy stretchblt*/
static void
FB7k_stretchblit(PSD dstpsd, MWCOORD dstx, MWCOORD dsty, MWCOORD dstw,
	MWCOORD dsth, PSD srcpsd, MWCOORD srcx, MWCOORD srcy, MWCOORD srcw,
	MWCOORD srch, long op)
{
	ADDR8	dst;
	ADDR8	src;
	int	dlinelen = dstpsd->linelen;
	int	slinelen = srcpsd->linelen;
	int	i, ymax;
	int	row_pos, row_inc;
	int	col_pos, col_inc;
	unsigned char pixel = 0;

	assert (dstpsd->addr != 0);
	assert (dstx >= 0 && dstx < dstpsd->xres);
	assert (dsty >= 0 && dsty < dstpsd->yres);
	assert (dstw > 0);
	assert (dsth > 0);
	assert (srcpsd->addr != 0);
	assert (srcx >= 0 && srcx < srcpsd->xres);
	assert (srcy >= 0 && srcy < srcpsd->yres);
	assert (srcw > 0);
	assert (srch > 0);
	assert (dstx+dstw <= dstpsd->xres);
	assert (dsty+dsth <= dstpsd->yres);
	assert (srcx+srcw <= srcpsd->xres);
	assert (srcy+srch <= srcpsd->yres);

	DRAWON;
	row_pos = 0x10000;
	row_inc = (srch << 16) / dsth;

	/* stretch blit using integer ratio between src/dst height/width*/
	for (ymax = dsty+dsth; dsty<ymax; ++dsty) {

		/* find source y position*/
		while (row_pos >= 0x10000L) {
			++srcy;
			row_pos -= 0x10000L;
		}

		dst = dstpsd->addr + dstx + dsty*dlinelen;
		src = srcpsd->addr + srcx + (srcy-1)*slinelen;

		/* copy a row of pixels*/
		col_pos = 0x10000;
		col_inc = (srcw << 16) / dstw;
		for (i=0; i<dstw; ++i) {
			/* get source x pixel*/
			while (col_pos >= 0x10000L) {
				pixel = *src++;
				col_pos -= 0x10000L;
			}
			*dst++ = pixel;
			col_pos += col_inc;
		}

		row_pos += row_inc;
	}
	DRAWOFF;
}

/* kaben's linedraw */
/* XXX 2005-12-08 aaron
 * does this draw the last pixel or not? */
static void
FB7k_conditioned_line(PSD psd,
    MWCOORD x1, MWCOORD y1, MWCOORD x2, MWCOORD y2,
    MWCOORD dx, MWCOORD dy, MWCOORD xi, MWCOORD yi,
    MWCOORD r, MWPIXELVAL c)
{
	extern unsigned long gr_dashmask;
	extern unsigned long gr_dashcount;
	extern char gr_dashindex;
	extern int gr_linestyle;

	ADDR8 addr = psd->addr;
	int linelen = psd->linelen;
	int dM, dm, adM, adm, M;

	assert (addr != 0);
	assert (addr <= (ADDR8)((ADDR8)psd->addr + psd->size));
	assert (addr >= (ADDR8)psd->addr);
	assert (x1 >= 0 && x1 < psd->xres);
	assert (x2 >= 0 && x2 < psd->xres);
	assert (y1 >= 0 && y1 < psd->yres);
	assert (y2 >= 0 && y2 < psd->yres);
	assert (c < psd->ncolors);

	DRAWON;
	addr += x1 + y1 * linelen;

	if (dx > dy){
		dM = dy;
		dm = dx - dM;
		M = xi * (x2 - x1);
		adM = xi;
		adm = adM + yi * linelen;
	} else {
		dM = dx;
		dm = dy - dM;
		M = yi * (y2 - y1);
		adM = yi * linelen;
		adm = adM + xi;
	}

	if((gr_dashmask == 0) || (gr_linestyle == MWLINE_SOLID)) {
		/* generic bresenham */
		while(M--) {
#if 0
			assert (addr <= (ADDR8)((ADDR8)psd->addr + psd->size));
			assert (addr >= (ADDR8)psd->addr);
#endif
			*addr = c;
			if (r >= dm) { r -= dm; addr += adm; }
			else { r += dM; addr += adM; }
		}
	} else while(M--) {
		/* bresenham with dashmask */
		if(gr_dashmask & (1 << gr_dashindex)) {
#if 0
			assert (addr <= (ADDR8)((ADDR8)psd->addr + psd->size));
			assert (addr >= (ADDR8)psd->addr);
#endif
			*addr = c;
		}
		gr_dashindex++;
		gr_dashindex%=gr_dashcount;
		if (r >= dm) {
			r -= dm;
		 	addr += adm;
		} else {
			r += dM;
		 	addr += adM;
		}
	}

	DRAWOFF;
}

/* Initialize memory device with passed parms, and select framebuffer
 * subdriver... */
MWBOOL
FB7k_mapmemgc(PSD mempsd,MWCOORD w,MWCOORD h,int planes,int bpp,int
    linelen, int size,void *addr)
{
	/* allocate memory */
	initmemgc(mempsd, w, h, planes, bpp, linelen, size, addr);

	/* fill in subdriver goo */
	mempsd->DrawPixel 	= FB7k_subdriver.DrawPixel;
	mempsd->ReadPixel	= FB7k_subdriver.ReadPixel;
	mempsd->DrawHorzLine 	= FB7k_subdriver.DrawHorzLine;
	mempsd->DrawVertLine 	= FB7k_subdriver.DrawVertLine;
	mempsd->FillRect 	= FB7k_subdriver.FillRect;
	mempsd->Blit 		= FB7k_subdriver.Blit;
	mempsd->DrawArea	= FB7k_subdriver.DrawArea;
	mempsd->StretchBlit 	= FB7k_subdriver.StretchBlit;
	mempsd->Line		= FB7k_subdriver.Line;
#if 0
	mempsd->DrawText	= FB7k_DrawText;
#else
	/* XXX FIXME XXX */
	if (mempsd->Blit == FB7kBitBLT_blit)
	{
	    mempsd->DrawText = FB7kBitBLT_DrawText;
	}
	else
	{
	    mempsd->DrawText = FB7k_DrawText;
	}
#endif

	/* initialize it */
	if(!FB7k_subdriver.Init(mempsd)) return FALSE;

	return TRUE;
}

/* hardware-specific text draw. */
static void FB7k_DrawText(PSD psd,MWCOORD x,MWCOORD y,
    const MWUCHAR *str, int count, MWPIXELVAL fg, PMWFONT pfont)
{
#if 1
	/* low-overhead byte-accessed with early bomb.
	 * currently the fastest of the bunch. */
	extern MWPIXELVAL gr_foreground;
	ADDR8	startcel = psd->addr + x + y * psd->linelen;

	/* loop through the count of characters */
	while (--count >= 0) {
		MWUCHAR			ch = *str++;
		MWCOORD			width, height, base;
		const MWIMAGEBITS	*charbitmap;
		ADDR8	startline = startcel;

		/* get the bitmap for the current character */
		pfont->fontprocs->GetTextBits(pfont, ch, &charbitmap,
		    &width, &height, &base);

		while (height--) {
			/* read a new row from our bitmap */
			MWIMAGEBITS	charrow = *charbitmap++;
			MWCOORD curwidth = width;
			ADDR8 curpixel = startline;

			/* write out bytewise if there's anything left */
			while (curwidth-- > 0 && charrow != 0) {
				if (MWIMAGE_TESTBIT(charrow))
					*curpixel++ = gr_foreground;
				else curpixel++;
				charrow = MWIMAGE_SHIFTBIT(charrow);
			}
			/* go to the next line */
			startline += psd->linelen;
		}

		/* head over to the next character */
		startcel += width;
	}
#else
	/* nibble to 32-bit lookup.  this is somewhat evil because we're
	 * directly abusing a priori knowledge of what the bitmap format
	 * is, but god knows we need the speed... */
	extern MWPIXELVAL gr_foreground;
	extern MWPIXELVAL gr_background;
	ADDR8	startcel = psd->addr + x + y * psd->linelen;
	/* nibble to unsigned long expansion */
	const unsigned long	charlut[16] = {
	    0x00000000, 0x000000ff, 0x0000ff00, 0x0000ffff,
	    0x00ff0000, 0x00ff00ff, 0x00ffff00, 0x00ffffff,
	    0xff000000, 0xff0000ff, 0xff00ff00, 0xff00ffff,
	    0xffff0000, 0xffff00ff, 0xffffff00, 0xffffffff
	};
	unsigned long fullforeground, fullbackground;

	/* only need to calculate these once per call */
	fullforeground =  gr_foreground&0xff;
	fullforeground |= fullforeground<<8;
	fullforeground |= fullforeground<<16;
	
	fullbackground =  gr_background&0xff;
	fullbackground |= fullbackground<<8;
	fullbackground |= fullbackground<<16;

	while (--count >= 0) {
		char			ch = *str++;
		MWCOORD			width, height, base;
		const MWIMAGEBITS	*charbitmap;
		ADDR8			startline = startcel;

		/* get the bitmap for the current character */
		pfont->fontprocs->GetTextBits(pfont, ch, &charbitmap,
		    &width, &height, &base);

		assert (width >= sizeof(unsigned long) + 1);

		while (height--) {
			/* read a row from the bitmap */
			MWIMAGEBITS	charrow = *charbitmap++;
			MWCOORD		curwidth = width;
			ADDR8		curpixel = startline;

			/* anything on this row? */
			if (charrow == 0) goto nextrow;

			/* check for alignment */
			if ((unsigned long) curpixel & 1) {
				if (MWIMAGE_TESTBIT(charrow))
				    *curpixel++ = gr_foreground;
				else curpixel++;
				charrow = MWIMAGE_SHIFTBIT(charrow);
				curwidth--;
			}

			/* anything left in the next four pixels? */
			if ((charrow & 0xf000) != 0) {
				unsigned long	brushmask =
				    charlut[(charrow & 0xf000)>>12];
				unsigned long	brush32;

				brush32 = brushmask & fullforeground;
				/* I hate this check, but it's cheap */
				if (fullbackground != 0)
				    brush32 |= ~brushmask & fullbackground;

				*((unsigned long *)curpixel)++ = brush32;
				curwidth-=sizeof(unsigned long);
				charrow=MWIMAGE_SHIFTBIT(charrow);
				charrow=MWIMAGE_SHIFTBIT(charrow);
				charrow=MWIMAGE_SHIFTBIT(charrow);
				charrow=MWIMAGE_SHIFTBIT(charrow);
			}

			/* anything left? */
			if (charrow == 0) goto nextrow;

			/* write out bytewise whatever is left */
			while (curwidth-- > 0 && charrow != 0) {
				if (MWIMAGE_TESTBIT(charrow))
					*curpixel++ = gr_foreground;
				else curpixel++;
				charrow = MWIMAGE_SHIFTBIT(charrow);
			}
nextrow:
			/* go to the next line */
			startline += psd->linelen;
		}

		/* head over to the next character */
		startcel += width;
	}
#endif
}

/* hardware-specific text draw. */
static void FB7kBitBLT_DrawText(PSD psd, MWCOORD x, MWCOORD y,
    const MWUCHAR *str, int count, MWPIXELVAL fg, PMWFONT pfont)
{
	extern MWPIXELVAL gr_foreground, gr_background;
	ADDR8	startcel;

	/* see if we're in the right place to use this routine.  if not,
	 * fall back to the un-accelerated version. */
	if ((u_int)psd->addr < VID_MEMBASE
	    || (u_int)psd->addr >= VID_MEMBASE+VID_MEMSIZE) {
		return FB7k_DrawText(psd, x, y, str, count, fg, pfont);
	}

	/* calculate our starting cel */
	startcel = psd->addr - VID_MEMBASE + x + y * psd->linelen;

	/* sanity checks */
	assert((u_int)psd->addr >= VID_MEMBASE);
	assert((u_int)psd->addr <  VID_MEMBASE+VID_MEMSIZE);

	DRAWON;

	/* loop through the count of characters */
	while (--count >= 0) {
		MWUCHAR			ch = *str++;
		MWCOORD			width, curwidth, height,
					curheight, base;
		int			curstride;
		unsigned int		curpixel;
		const MWIMAGEBITS	*charbitmap;
		ADDR8			startline = startcel;

		/* get the bitmap for the current character */
		pfont->fontprocs->GetTextBits(pfont, ch, &charbitmap,
		    &width, &height, &base);

		/* read a new row from our bitmap */
		curwidth = width - 1;
		curheight = height - 1;
		curpixel = (u_int)startline;
		curstride = psd->linelen;

		VID506_REG(VID506_BITBLT_SRC_START_ADDR0) = 0;

		/* result is going here */
		VID506_REG(VID506_BITBLT_DEST_START_ADDR0) = curpixel & 0xff;
		curpixel>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR1) = curpixel & 0xff;
		curpixel>>=8;
		VID506_REG(VID506_BITBLT_DEST_START_ADDR2) = curpixel & 0xff;

		/* with this width */
		VID506_REG(VID506_BITBLT_WIDTH0) = curwidth & 0xff;
		curwidth>>=8;
		VID506_REG(VID506_BITBLT_WIDTH1) = curwidth & 0xff;

		/* and this height
		 * since fonts are encoded one line per word, this is
		 * also the number of words we will be feeding the
		 * bitblt engine.  */
		VID506_REG(VID506_BITBLT_HEIGHT0) = curheight & 0xff;
		curheight>>=8;
		VID506_REG(VID506_BITBLT_HEIGHT1) = curheight & 0xff;

		/* we are doing color expansion */
		VID506_REG(VID506_BITBLT_OPERATION) = 0x08;

		/* the ROP code is always 15 since we always start with
		 * the left edge of the source bits.  (big endian.)  if
		 * we were little endian, we'd start with bit 7. */
		VID506_REG(VID506_BITBLT_ROP_CODE) = 0x0f;

	    	/* background and foreground colors
		 * XXX assume 8bpp here, otherwise we need to write the
		 * COLOR1 components too. */
		VID506_REG(VID506_BITBLT_BACKGND_COLOR0) =
		    gr_background & 0xff;
		VID506_REG(VID506_BITBLT_FOREGND_COLOR0) =
		    gr_foreground & 0xff;

		/* we are 8bpp */
		VID506_REG(VID506_BITBLT_CTRL1) = 0;

		/* screen stride for destination */
		curstride>>=1;
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET0) = curstride & 0xff;
		curstride>>=8;
		VID506_REG(VID506_BITBLT_MEM_ADDR_OFFSET1) = curstride & 0xff;

		/* gentlemen, start your BLT engines */
		VID506_REG(VID506_BITBLT_CTRL0) = 0x80;

		/* wait for blit engine to start */
		while ((VID506_REG(VID506_BITBLT_CTRL0) & 0x80) == 0);

		/* loop over the words to send */
		while(height--)
		{
			MWIMAGEBITS charrow = *charbitmap++;
			unsigned char fifostat;

			/* wait for half full and FIFO full flags to become
			 * empty */
			do
			{
		    		fifostat =
				    VID506_REG(VID506_BITBLT_CTRL0);
			} while ((fifostat & 0x30) != 0);

			/* we know there is room in the FIFO for at
			 * least one write.  if we have at least two
			 * lines to write, and there is space, do both
			 * the writes.  */
			if (height > 1 && (fifostat & 0x40) == 0)
			{
	    			/* two writes to FIFO */
    				VID506_BLTMEM16(0) = charrow;
				charrow = *charbitmap++;
				VID506_BLTMEM16(2) = charrow;
				height--;
			}
			else
			{
    				/* one write to FIFO */
    				VID506_BLTMEM16(0) = charrow;
			}
		}

		/* make sure the bitblt is completely finished before
		 * proceeding */
		BitBLTspin();

		/* head over to the next character */
		startcel += width;
	}

	DRAWOFF;
}

static PSD
FB7k_open(PSD psd)
{
    vidchip_e chiptype;
    char *modestring;
    modeline_t modeline;
    int iclk = (VID_ICLK * 1000.0);
    int result;

    /* attempt to detect a chip */
    chiptype = vid_detect();
    if (chiptype == -1) return NULL;

    dprintf("iclk: %d\n", iclk);

    /* found a chip; pick a modeline based on which iclk we have */
    switch (iclk)
    {
	case 25175:
	    modestring = MODELINE25175;
	    break;
	case 31500:
	    modestring = MODELINE31500;
	    break;
	case 32000:
	    modestring = MODELINE32000;
	    break;
	case 40000:
	    modestring = MODELINE40000;
	    break;
	default:
	    dprintf("no modeline\n");
	    return NULL;
    }

    /* parse down the modeline */
    result = parse_modeline(modestring, &modeline, SCREEN_DEPTH);
    if (result != 0) return NULL;

    switch (chiptype) {

    case S1D13505:
	result = vid505_setmode(&modeline, SCREEN_DEPTH);
	if (result != 0) return NULL;
	scrdev.SetPalette = FB7k505_setpalette;
	break;

    case S1D13506:
    	result = vid506_setmode(&modeline, SCREEN_DEPTH);
	if (result != 0) return NULL;
	scrdev.SetPalette = FB7k506_setpalette;
	FB7k_subdriver.FillRect = FB7kBitBLT_drawfillrect;
	FB7k_subdriver.Blit = FB7kBitBLT_blit;
	break;

    default:
    	return NULL;
	break;

    }

    /* init driver variables depending on ega/vga mode*/
    psd->xres = psd->xvirtres = modeline.h_res;
    psd->yres = psd->yvirtres = modeline.v_res;

    switch(SCREEN_DEPTH){
    case 1: psd->linelen = psd->xres/8; break;
    case 2: psd->linelen = psd->xres/4; break;
    case 4: psd->linelen = psd->xres/2; break;
    case 8: psd->linelen = psd->xres; break;
    case 15: case 16: psd->linelen = psd->xres * 2; break;
    default:
        /* unsupported screen depth */
        return NULL;
        break;
    }
    psd->planes = 1;
    psd->bpp = SCREEN_DEPTH;
    psd->pixtype = MWPIXEL_FORMAT;
    psd->ncolors = psd->bpp >= 24? (1 << 24) : (1 << psd->bpp);
    psd->flags = PSF_SCREEN | PSF_HAVEBLIT;

    /* set and initialize subdriver into screen driver */
    psd->DrawPixel 	= FB7k_subdriver.DrawPixel;
    psd->ReadPixel	= FB7k_subdriver.ReadPixel;
    psd->DrawHorzLine 	= FB7k_subdriver.DrawHorzLine;
    psd->DrawVertLine 	= FB7k_subdriver.DrawVertLine;
    psd->FillRect 	= FB7k_subdriver.FillRect;
    psd->Blit 		= FB7k_subdriver.Blit;
    psd->DrawArea	= FB7k_subdriver.DrawArea;
    psd->StretchBlit 	= FB7k_subdriver.StretchBlit;
    psd->Line		= FB7k_subdriver.Line;
    /* XXX FIXME XXX */
    if (psd->Blit == FB7kBitBLT_blit)
    {
	psd->DrawText = FB7kBitBLT_DrawText;
    }
    else
    {
	psd->DrawText = FB7k_DrawText;
    }

    /* run init */
    if(!FB7k_subdriver.Init(psd)) {
        dprintf("<FB7k_open> Driver initialize failed.\n");
        return NULL;
    }

    dprintf("<FB7k_open>\n");
    return psd;
}

static void
FB7k_close(PSD psd)
{
    dprintf("<FB7k_close>\n");
}

static void
FB7k_getscreeninfo(PSD psd,PMWSCREENINFO psi)
{
    psi->rows = psd->yvirtres;
    psi->cols = psd->xvirtres;
    psi->planes = psd->planes;
    psi->bpp = psd->bpp;
    psi->ncolors = psd->ncolors;
    psi->pixtype = psd->pixtype;
    psi->fonts = NUMBER_FONTS;

    /* FB7k 640x480*/
    psi->xdpcm = 27;    /* assumes screen width of 24 cm*/
    psi->ydpcm = 27;    /* assumes screen height of 18 cm*/
    ddprintf("<FB7k_getscreeninfo>\n");
}

/* set the palette
 *
 * note that for the S1D13505 chip used in the 7000, only the top nibble
 * is relevant for color selection.  rather than export this limitation
 * to the bottom layers, we leave it quantized.  */
static void
FB7k505_setpalette(PSD psd, int first, int count, MWPALENTRY *pal) {
#if SCREEN_DEPTH <= 8
    int palindex;
    dprintf("<FB7k505_setpalette> entered.\n");
    dprintf(
    "<FB7k505_setpalette> initial values: first %d, count %d, FB7k_pal_max %d.\n",
    first, count, FB7k_pal_max
    );

	/* FIXME:  Shouldn't fail silently.  Must give error handling. */
	if (first + count > FB7k_pal_max) {
		dprintf("<FB7k505_setpalette> dies.\n");
		return;
	}

	dprintf("<FB7k505_setpalette> ");
	for (palindex = 0; palindex < count; palindex++) {
		VID505_REG(VID505_LUT_ADDR) = first + palindex;
		VID505_REG(VID505_LUT_DATA) = pal[palindex].r;
		VID505_REG(VID505_LUT_DATA) = pal[palindex].g;
		VID505_REG(VID505_LUT_DATA) = pal[palindex].b;
		dprintf("%3d: %02x %02x %02x\t", first + palindex,
		    pal[palindex].r, pal[palindex].g, pal[palindex].b);
		if (palindex > 0 && palindex%3 == 0)
		    dprintf("\n<FB7k505_setpalette> ");
	}
	dprintf("\n");
#else
	dprintf("<FB7k505_setpalette> not used in 8+bpp\n");
#endif
}

/* set the palette
 *
 * S1D13506 has same caveats for palette as the 505... */
static void
FB7k506_setpalette(PSD psd, int first, int count, MWPALENTRY *pal) {
#if SCREEN_DEPTH <= 8
    int palindex;
    ddprintf("<FB7k506_setpalette> entered.\n");
    ddprintf(
    "<FB7k506_setpalette> initial values: first %d, count %d, FB7k_pal_max %d.\n",
    first, count, FB7k_pal_max
    );

	/* FIXME:  Shouldn't fail silently.  Must give error handling. */
	if (first + count > FB7k_pal_max) {
		dprintf("<FB7k506_setpalette> dies.\n");
		return;
	}

	ddprintf("<FB7k506_setpalette> ");
	for (palindex = 0; palindex < count; palindex++) {
		VID506_REG(VID506_LUT_ADDR) = first + palindex;
		VID506_REG(VID506_LUT_DATA) = pal[palindex].r;
		VID506_REG(VID506_LUT_DATA) = pal[palindex].g;
		VID506_REG(VID506_LUT_DATA) = pal[palindex].b;
		ddprintf("%3d: %02x %02x %02x\t", first + palindex,
		    pal[palindex].r, pal[palindex].g, pal[palindex].b);
		if (palindex > 0 && palindex%3 == 0)
		    ddprintf("\n<FB7k506_setpalette> ");
	}
	ddprintf("\n");
#else
	ddprintf("<FB7k506_setpalette> not used in 8+bpp\n");
#endif
}
